@@ -5,17 +5,30 @@ from .exceptions import (
     InvalidInputType,
     InvalidInputValue,
     MissingInputs,
-    ServerError,
     ServiceError,
-    ThreadingException,
+    ServiceUnavailable,
     ZeroMatched,
 )
 from .print_versions import show_versions
 from .pygeoogc import WFS, WMS, ArcGISRESTful, ServiceURL
-from .utils import MatchCRS, RetrySession
+from .utils import RetrySession
 
 try:
     __version__ = get_distribution(__name__).version
 except DistributionNotFound:
-    # package is not installed
-    pass
+    __version__ = "999"
+
+__all__ = [
+    "ArcGISRESTful",
+    "WFS",
+    "WMS",
+    "ServiceURL",
+    "RetrySession",
+    "InvalidInputType",
+    "InvalidInputValue",
+    "MissingInputs",
+    "ServiceError",
+    "ServiceUnavailable",
+    "ZeroMatched",
+    "show_versions",
+]
