@@ -3,9 +3,12 @@ import contextlib
 import logging
 import sys
 import urllib.parse as urlparse
+import uuid
 from dataclasses import dataclass
-from typing import Dict, List, Optional, Tuple, Union
+from pathlib import Path
+from typing import Any, Dict, List, Optional, Tuple, Union
 
+import async_retriever as ar
 import cytoolz as tlz
 import pyproj
 from owslib.wfs import WebFeatureService
@@ -19,11 +22,10 @@ from .exceptions import (
     InvalidInputType,
     InvalidInputValue,
     MissingInputs,
-    ServerError,
     ServiceError,
+    ServiceUnavailable,
     ZeroMatched,
 )
-from .utils import RetrySession
 
 logger = logging.getLogger(__name__)
 logger.setLevel(logging.INFO)
@@ -66,7 +68,7 @@ class RESTValidator(BaseModel):
     outformat: str = "geojson"
     outfields: Union[List[str], str] = "*"
     crs: str = DEF_CRS
-    max_workers: int = 2
+    max_workers: int = 4
 
     @validator("base_url")
     def _layer_from_url(cls, v):
@@ -90,8 +92,8 @@ class RESTValidator(BaseModel):
     def _valid_crs(cls, v):
         try:
             return pyproj.CRS(v)
-        except pyproj.exceptions.CRSError:
-            raise InvalidInputType("crs", "a valid CRS")
+        except pyproj.exceptions.CRSError as ex:
+            raise InvalidInputType("crs", "a valid CRS") from ex
 
     @validator("max_workers")
     def _positive_integer_threads(cls, v):
@@ -175,8 +177,11 @@ class ArcGISRESTfulBase:
         self.valid_fields: List[str] = []
         self.field_types: Dict[str, str] = {}
         self.feature_types: Optional[Dict[str, str]] = None
-
-        self.session = RetrySession()
+        self.retry: bool = False
+        self.return_m: bool = False
+        self.n_missing: int = 0
+        self.total_n_features: int = 0
+        self.failed_path: Path = Path("cache", f"{uuid.uuid4().hex}.txt")
 
         self.initialize_service()
 
@@ -203,25 +208,53 @@ class ArcGISRESTfulBase:
         oid_list = [str(oids)] if isinstance(oids, int) else [str(v) for v in oids]
 
         self.n_features = len(oid_list)
-        logger.info(f"Found {self.n_features:,} features in the requested region.")
-        return list(tlz.partition_all(self.max_nrecords, oid_list))
+        if not self.retry:
+            logger.info(f"Found {self.n_features:,} features in the requested region.")
+        return list(tlz.partition_all(self.max_nrecords, sorted(oid_list)))
+
+    def get_features(self, return_m: bool = False) -> List[Dict[str, Any]]:
+        """Get features based on the feature IDs.
+
+        Parameters
+        ----------
+        return_m : bool
+            Whether to activate the Return M (measure) in the request, defaults to False.
+
+        Returns
+        -------
+        dict
+            (Geo)json response from the web service.
+        """
+        payloads = [
+            {
+                "objectIds": ",".join(ids),
+                "returnGeometry": "true",
+                "outSR": self.out_sr,
+                "outfields": ",".join(self.outfields),
+                "ReturnM": f"{return_m}".lower(),
+                "f": self.outformat,
+            }
+            for ids in self.featureids
+        ]
+
+        return self._get_response(payloads, "POST")
 
     def _set_service_properties(self) -> None:
-        resp = self.session.get(self.base_url, {"f": "json"})
-        utils.check_response(resp)
-
         try:
-            rjson = resp.json()
+            rjson = ar.retrieve([self.base_url], "json", [{"params": {"f": "json"}}])[0]
             self.valid_layers = {f"{lyr['id']}": lyr["name"] for lyr in rjson["layers"]}
             self.query_formats = rjson["supportedQueryFormats"].replace(" ", "").lower().split(",")
 
             extent = rjson["extent"] if "extent" in rjson else rjson["fullExtent"]
             bounds = (extent["xmin"], extent["ymin"], extent["xmax"], extent["ymax"])
-            crs = extent["spatialReference"]["latestWkid"]
-            self.extent = utils.MatchCRS(crs, DEF_CRS).bounds(bounds)
+            for f in ["latestWkid", "wkid", "wkt"]:
+                if f in extent["spatialReference"]:
+                    crs = extent["spatialReference"][f]
+                    break
+            self.extent = utils.match_crs(bounds, crs, DEF_CRS)
 
-        except (JSONDecodeError, KeyError):
-            raise ServerError(self.base_url)
+        except (JSONDecodeError, KeyError) as ex:
+            raise ServiceError(self.base_url) from ex
 
         with contextlib.suppress(KeyError):
             self.units = rjson["units"].replace("esri", "").lower()
@@ -229,10 +262,7 @@ class ArcGISRESTfulBase:
 
     def _set_layer_properties(self) -> None:
         """Set properties of the target layer."""
-        resp = self.session.get(self.url, {"f": "json"})
-        utils.check_response(resp)
-        rjson = resp.json()
-
+        rjson = ar.retrieve([self.url], "json", [{"params": {"f": "json"}}])[0]
         self.valid_fields = list(
             set(
                 utils.traverse_json(rjson, ["fields", "name"])
@@ -261,29 +291,120 @@ class ArcGISRESTfulBase:
         geo_crs: str = DEF_CRS,
     ) -> Dict[str, Union[str, bytes]]:
         """Generate geometry queries based on ESRI template."""
-        match_crs = utils.MatchCRS(geo_crs, self.crs)
+        geom = utils.match_crs(geom, geo_crs, self.crs)
 
         if isinstance(geom, tuple) and len(geom) == 4:
-            geom = match_crs.bounds(geom)  # type: ignore
             return utils.ESRIGeomQuery(geom, self.out_sr).bbox()
 
         if isinstance(geom, Point):
-            geom = match_crs.geometry(geom)
             return utils.ESRIGeomQuery((geom.x, geom.y), self.out_sr).point()
 
         if isinstance(geom, MultiPoint):
-            geom = match_crs.geometry(geom)
             return utils.ESRIGeomQuery([(g.x, g.y) for g in geom], self.out_sr).multipoint()
 
         if isinstance(geom, Polygon):
-            geom = match_crs.geometry(geom)
             return utils.ESRIGeomQuery(geom, self.out_sr).polygon()
 
         if isinstance(geom, LineString):
-            geom = match_crs.geometry(geom)
             return utils.ESRIGeomQuery(geom, self.out_sr).polyline()
 
-        raise InvalidInputType("geom", "LineString, Polygon, Point, MultiPoint, tuple, list")
+        raise InvalidInputType("geom", "LineString, Polygon, Point, MultiPoint, tuple")
+
+    def _retry(self, return_m: bool, partition_fac: float) -> List[Dict[str, Any]]:
+        """Retry failed requests."""
+        with open(self.failed_path) as f:
+            oids = [int(i) for i in f.read().splitlines()]
+
+        max_nrecords = self.max_nrecords
+        self.max_nrecords = int(partition_fac * max_nrecords)
+        self.featureids = self.partition_oids(oids)
+        self.max_nrecords = max_nrecords
+
+        return self.get_features(return_m)
+
+    def _retry_failed_requests(self) -> List[Dict[str, Any]]:
+        """Retry failed requests."""
+        self.retry = True
+        partition_fac = 0.5
+        features = []
+        while True:
+            try:
+                features.append(self._retry(self.return_m, partition_fac))
+            except ServiceError:
+                partition_fac *= 0.5
+                try:
+                    features.append(self._retry(self.return_m, partition_fac))
+                except ServiceError:
+                    features.append(self._retry(self.return_m, 1.0 / self.max_nrecords))
+                    break
+
+        logger.info(
+            f"Total of {self.n_missing} out of {self.total_n_features} "
+            + "features are not available on the server."
+        )
+        self.failed_path.unlink()
+        self.retry = False
+        return list(tlz.concat(features))
+
+    def _get_response(
+        self, payloads: List[Dict[str, str]], method: str = "GET"
+    ) -> List[Dict[str, Any]]:
+        """Send payload and get the response."""
+        req_key = "params" if method == "GET" else "data"
+        urls, kwds = zip(*((self.query_url, {req_key: p}) for p in payloads))
+        try:
+            resp = ar.retrieve(
+                urls, "json", kwds, request_method=method, max_workers=self.max_workers
+            )
+        except JSONDecodeError:
+            raise ZeroMatched
+
+        resp = self._cleanup_resp(resp, payloads)
+        return resp
+
+    def _cleanup_resp(
+        self, resp: List[Dict[str, Any]], payloads: List[Dict[str, str]]
+    ) -> List[Dict[str, Any]]:
+        """Remove failed responses."""
+        fails = [i for i, r in enumerate(resp) if "error" in r]
+
+        if len(fails) > 0:
+            err = resp[fails[0]]["error"]["message"]
+            resp = [r for i, r in enumerate(resp) if i not in fails]
+            if len(resp) == 0 and self.retry:
+                raise ServiceError(err)
+
+            if "objectIds" in payloads[fails[0]]:
+                oids = list(tlz.concat(payloads[i]["objectIds"].split(",") for i in fails))
+                self.n_missing = len(oids)
+
+                with open(self.failed_path, "w") as f:
+                    f.write("\n".join(oids))
+
+                if not self.retry:
+                    self.return_m = bool(payloads[0]["ReturnM"])
+                    self.total_n_features = self.n_features
+                    logger.info(
+                        " ".join(
+                            [
+                                f"Found {len(fails)} failed requests.",
+                                "Retrying to pluck out all available features.",
+                            ]
+                        )
+                    )
+                    resp.extend(self._retry_failed_requests())
+            else:
+                if len(resp) == 0:
+                    raise ZeroMatched
+
+                msg = (
+                    "Service failed to process some of the queries with"
+                    + "the following message, returning the rest:\n"
+                    + err
+                )
+                logger.warning(msg)
+
+        return resp
 
     def __repr__(self) -> str:
         """Print the service configuration."""
@@ -346,8 +467,8 @@ class WMSBase:
         """Validate input arguments with the WMS service."""
         try:
             wms = WebMapService(self.url, version=self.version)
-        except AttributeError:
-            raise ServiceError(self.url)
+        except AttributeError as ex:
+            raise ServiceUnavailable(self.url) from ex
 
         layers = [self.layers] if isinstance(self.layers, str) else self.layers
         valid_layers = {wms[lyr].name: wms[lyr].title for lyr in list(wms.contents)}
@@ -365,7 +486,10 @@ class WMSBase:
 
     def get_validlayers(self) -> Dict[str, str]:
         """Get the layers supported by the WMS service."""
-        wms = WebMapService(self.url, version=self.version)
+        try:
+            wms = WebMapService(self.url, version=self.version)
+        except AttributeError as ex:
+            raise ServiceUnavailable(self.url) from ex
 
         return {wms[lyr].name: wms[lyr].title for lyr in list(wms.contents)}
 
@@ -386,11 +510,14 @@ class WFSBase:
         The data format to request for data from the service, defaults to None which
          throws an error and includes all the available format offered by the service.
     version : str, optional
-        The WFS service version which should be either 1.1.1, 1.3.0, or 2.0.0.
+        The WFS service version which should be either 1.0.0, 1.1.0, or 2.0.0.
         Defaults to 2.0.0.
     crs: str, optional
         The spatial reference system to be used for requesting the data, defaults to
         epsg:4326.
+    read_method : str, optional
+        Method for reading the retrieved data, defaults to ``json``. Valid options are
+        ``json``, ``binary``, and ``text``.
     """
 
     url: str
@@ -398,24 +525,28 @@ class WFSBase:
     outformat: Optional[str] = None
     version: str = "2.0.0"
     crs: str = DEF_CRS
+    read_method: str = "json"
 
     def __repr__(self) -> str:
         """Print the services properties."""
-        return (
-            "Connected to the WFS service with the following properties:\n"
-            + f"URL: {self.url}\n"
-            + f"Version: {self.version}\n"
-            + f"Layer: {self.layer}\n"
-            + f"Output Format: {self.outformat}\n"
-            + f"Output CRS: {self.crs}"
+        return "\n".join(
+            [
+                "Connected to the WFS service with the following properties:",
+                f"URL: {self.url}",
+                f"Version: {self.version}",
+                f"Layer: {self.layer}",
+                f"Output Format: {self.outformat}",
+                f"Output CRS: {self.crs}",
+                f"Read Method: {self.read_method}",
+            ]
         )
 
     def validate_wfs(self) -> None:
         """Validate input arguments with the WFS service."""
         try:
             wfs = WebFeatureService(self.url, version=self.version)
-        except AttributeError:
-            raise ServiceError(self.url)
+        except AttributeError as ex:
+            raise ServiceUnavailable(self.url) from ex
 
         valid_layers = list(wfs.contents)
         if self.layer is None:
@@ -424,11 +555,15 @@ class WFSBase:
                 + " The following layers are available:\n"
                 + ", ".join(valid_layers)
             )
-
         if self.layer not in valid_layers:
             raise InvalidInputValue("layers", valid_layers)
 
-        valid_outformats = wfs.getOperationByName("GetFeature").parameters["outputFormat"]["values"]
+        wfs_features = wfs.getOperationByName("GetFeature")
+        if self.version == "1.0.0":
+            valid_outformats = [f.rsplit("}", 1)[-1] for f in wfs_features.formatOptions]
+        else:
+            valid_outformats = wfs_features.parameters["outputFormat"]["values"]
+
         valid_outformats = [v.lower() for v in valid_outformats]
         if self.outformat is None:
             raise MissingInputs(
@@ -457,20 +592,16 @@ class WFSBase:
             max_features: 1,
         }
 
-        session = RetrySession()
-        resp = session.get(self.url, payload)
-        utils.check_response(resp)
-
-        r_json = resp.json()
+        rjson = ar.retrieve([self.url], "json", [{"params": payload}])[0]
         valid_fields = list(
             set(
-                utils.traverse_json(r_json, ["fields", "name"])
-                + utils.traverse_json(r_json, ["fields", "alias"])
+                utils.traverse_json(rjson, ["fields", "name"])
+                + utils.traverse_json(rjson, ["fields", "alias"])
                 + ["*"]
             )
         )
 
         if None in valid_fields:
-            valid_fields = list(utils.traverse_json(r_json, ["features", "properties"])[0].keys())
+            valid_fields = list(utils.traverse_json(rjson, ["features", "properties"])[0].keys())
 
         return valid_fields
