@@ -1,5 +1,4 @@
 """Base classes and function for REST, WMS, and WMF services."""
-import collections
 import itertools
 import logging
 import sys
@@ -7,18 +6,15 @@ from pathlib import Path
 from types import SimpleNamespace
 from typing import Any, Dict, List, Optional, Tuple, Union
 
-import cytoolz as tlz
+import async_retriever as ar
 import pyproj
 import shapely.ops as ops
 import yaml
-from requests import Response
 from shapely.geometry import LineString, MultiPoint, MultiPolygon, Point, Polygon
-from simplejson import JSONDecodeError
 
 from . import utils
 from .core import ArcGISRESTfulBase, WFSBase, WMSBase
 from .exceptions import InvalidInputType, InvalidInputValue, ZeroMatched
-from .utils import MatchCRS, RetrySession
 
 logger = logging.getLogger(__name__)
 logger.setLevel(logging.INFO)
@@ -77,9 +73,9 @@ class ArcGISRESTful(ArcGISRESTfulBase):
         Parameters
         ----------
         geom : LineString, Polygon, Point, MultiPoint, tuple, or list of tuples
-            A geometry (LineString, Polygon, Point, MultiPoint), tuple of length
-             2 (``(x, y)``), a list of tuples of length 2 (``[(x, y), ...]``), or bounding box
-             (tuple of length 4 (``(xmin, ymin, xmax, ymax)``)).
+            A geometry (LineString, Polygon, Point, MultiPoint), tuple of length two
+            (``(x, y)``), a list of tuples of length 2 (``[(x, y), ...]``), or bounding box
+            (tuple of length 4 (``(xmin, ymin, xmax, ymax)``)).
         geo_crs : str
             The spatial reference of the input geometry, defaults to EPSG:4326.
         spatial_relation : str, optional
@@ -98,17 +94,10 @@ class ArcGISRESTful(ArcGISRESTfulBase):
             * ``esriSpatialRelRelation``
 
         sql_clause : str, optional
-            A valid SQL 92 WHERE clause, default to None.
+            Valid SQL 92 WHERE clause, default to None.
         distance : int, optional
-            The buffer distance in meters for the input geometries in meters, default to None.
+            Buffer distance in meters for the input geometries, default to None.
         """
-        if isinstance(geom, tuple) and len(geom) == 2:
-            geom = Point(geom)
-        elif isinstance(geom, list) and all(len(g) == 2 for g in geom):
-            geom = MultiPoint(geom)
-
-        geom_query = self._esri_query(geom, geo_crs)
-
         valid_spatialrels = [
             "esriSpatialRelIntersects",
             "esriSpatialRelContains",
@@ -123,6 +112,13 @@ class ArcGISRESTful(ArcGISRESTfulBase):
         if spatial_relation not in valid_spatialrels:
             raise InvalidInputValue("spatial_relation", valid_spatialrels)
 
+        if isinstance(geom, tuple) and len(geom) == 2:
+            geom = Point(geom)
+        elif isinstance(geom, list) and all(len(g) == 2 for g in geom):
+            geom = MultiPoint(geom)
+
+        geom_query = self._esri_query(geom, geo_crs)
+
         payload = {
             **geom_query,  # type: ignore
             "spatialRel": spatial_relation,
@@ -136,14 +132,11 @@ class ArcGISRESTful(ArcGISRESTfulBase):
         if sql_clause:
             payload.update({"where": sql_clause})
 
-        resp = self.session.post(self.query_url, payload)
-
+        resp = self._get_response([payload], method="POST")[0]
         try:
-            self.featureids = self.partition_oids(resp.json()["objectIds"])
-        except JSONDecodeError:
-            raise ZeroMatched
-        except KeyError:
-            raise ZeroMatched(f"Service error message:\n{resp.json()['error']['message']}")
+            self.featureids = self.partition_oids(resp["objectIds"])
+        except KeyError as ex:
+            raise ZeroMatched(resp["error"]["message"]) from ex
 
     def oids_byfield(self, field: str, ids: Union[str, List[str]]) -> None:
         """Get Object IDs based on a list of field IDs.
@@ -188,66 +181,12 @@ class ArcGISRESTful(ArcGISRESTfulBase):
             "returnIdsOnly": "true",
             "f": self.outformat,
         }
-        resp = self.session.post(self.query_url, payload)
 
+        resp = self._get_response([payload])[0]
         try:
-            self.featureids = self.partition_oids(resp.json()["objectIds"])
-        except JSONDecodeError:
-            raise ZeroMatched
-        except KeyError:
-            raise ZeroMatched(f"Service error message:\n{resp.json()['error']['message']}")
-
-    def get_features(self, return_m: bool = False) -> List[Dict[str, Any]]:
-        """Get features based on the feature IDs.
-
-        Parameters
-        ----------
-        return_m : bool
-            Whether to activate the Return M (measure) in the request, defaults to False.
-
-        Returns
-        -------
-        dict
-            (Geo)json response from the web service.
-        """
-        payload = {
-            "returnGeometry": "true",
-            "outSR": self.out_sr,
-            "outfields": ",".join(self.outfields),
-            "ReturnM": return_m,
-            "f": self.outformat,
-        }
-
-        def getter(ids: Tuple[str, ...]) -> Union[Dict[str, Any], str]:
-            payload.update({"objectIds": ", ".join(ids)})
-            resp = self.session.post(self.query_url, payload)
-            try:
-                return resp.json()
-            except JSONDecodeError:
-                return "error"
-
-        resp = utils.threading(getter, self.featureids, max_workers=self.max_workers)
-
-        resp_types = collections.defaultdict(list)
-        for r in resp:
-            resp_types[type(r)].append(r)
-
-        valid_resp = resp_types.pop(dict)
-        if len(resp_types) != 0:
-            bad_resp_len = len(list(tlz.concat(resp_types.values())))
-            logger.warning(
-                ", ".join(
-                    [
-                        f"Found {bad_resp_len} errors in service responses",
-                        f"returning the {len(valid_resp)} remanining valid response(s).",
-                    ]
-                )
-            )
-
-        if len(valid_resp) == 0:
-            raise ZeroMatched
-
-        return valid_resp
+            self.featureids = self.partition_oids(resp["objectIds"])
+        except KeyError as ex:
+            raise ZeroMatched(resp["error"]["message"]) from ex
 
 
 class WMS(WMSBase):
@@ -285,7 +224,6 @@ class WMS(WMSBase):
     ) -> None:
         super().__init__(url, layers, outformat, version, crs)
 
-        self.session = RetrySession()
         self.layers = [self.layers] if isinstance(self.layers, str) else self.layers
         if validation:
             self.validate_wms()
@@ -297,6 +235,7 @@ class WMS(WMSBase):
         box_crs: str = DEF_CRS,
         always_xy: bool = False,
         max_px: int = 8000000,
+        kwargs: Optional[Dict[str, Any]] = None,
     ) -> Dict[str, bytes]:
         """Get data from a WMS service within a geometry or bounding box.
 
@@ -318,6 +257,9 @@ class WMS(WMSBase):
         max_px : int, opitonal
             The maximum allowable number of pixels (width x height) for a WMS requests,
             defaults to 8 million based on some trial-and-error.
+        kwargs: dict, optional
+            Optional additional keywords passed as payload, defaults to None.
+            For example, ``{"styles": "default"}``.
 
         Returns
         -------
@@ -326,7 +268,7 @@ class WMS(WMSBase):
             from the WMS service as bytes.
         """
         utils.check_bbox(bbox)
-        _bbox = MatchCRS(box_crs, self.crs).bounds(bbox)
+        _bbox = utils.match_crs(bbox, box_crs, self.crs)
         bounds = utils.bbox_decompose(_bbox, resolution, self.crs, max_px)
 
         payload = {
@@ -335,6 +277,12 @@ class WMS(WMSBase):
             "request": "GetMap",
         }
 
+        if not isinstance(kwargs, (dict, type(None))):
+            raise InvalidInputType("kwargs", "dict or None")
+
+        if isinstance(kwargs, dict):
+            payload.update(kwargs)
+
         if self.version == "1.1.1":
             payload["srs"] = self.crs
         else:
@@ -342,23 +290,29 @@ class WMS(WMSBase):
 
         geographic_crs = pyproj.CRS.from_user_input(self.crs).is_geographic
 
-        def _getmap(
+        def _get_payloads(
             args: Tuple[str, Tuple[Tuple[float, float, float, float], str, int, int]]
-        ) -> Tuple[str, bytes]:
+        ) -> Tuple[str, Dict[str, str]]:
             lyr, bnds = args
             _bbox, counter, _width, _height = bnds
 
             if self.version != "1.1.1" and geographic_crs and not always_xy:
                 _bbox = (_bbox[1], _bbox[0], _bbox[3], _bbox[2])
+            _payload = payload.copy()
+            _payload["bbox"] = f'{",".join(str(c) for c in _bbox)}'
+            _payload["width"] = str(_width)
+            _payload["height"] = str(_height)
+            _payload["layers"] = lyr
+            return f"{lyr}_dd_{counter}", _payload
 
-            payload["bbox"] = f'{",".join(str(c) for c in _bbox)}'
-            payload["width"] = str(_width)
-            payload["height"] = str(_height)
-            payload["layers"] = lyr
-            resp = self.session.get(self.url, payload)
-            return f"{lyr}_dd_{counter}", resp.content
-
-        return dict(_getmap(i) for i in itertools.product(self.layers, bounds))
+        layers, payloads = zip(*(_get_payloads(i) for i in itertools.product(self.layers, bounds)))
+        rbinary = ar.retrieve(
+            [self.url] * len(payloads),
+            "binary",
+            [{"params": p} for p in payloads],
+            max_workers=4,
+        )
+        return dict(zip(layers, rbinary))
 
 
 class WFS(WFSBase):
@@ -376,7 +330,7 @@ class WFS(WFSBase):
         The data format to request for data from the service, defaults to None which
          throws an error and includes all the available format offered by the service.
     version : str, optional
-        The WFS service version which should be either 1.1.1, 1.3.0, or 2.0.0.
+        The WFS service version which should be either 1.0.0, 1.1.0, or 2.0.0.
         Defaults to 2.0.0.
     crs: str, optional
         The spatial reference system to be used for requesting the data, defaults to
@@ -398,7 +352,6 @@ class WFS(WFSBase):
     ) -> None:
         super().__init__(url, layer, outformat, version, crs)
 
-        self.session = RetrySession()
         if validation:
             self.validate_wfs()
 
@@ -407,7 +360,7 @@ class WFS(WFSBase):
         bbox: Tuple[float, float, float, float],
         box_crs: str = DEF_CRS,
         always_xy: bool = False,
-    ) -> Response:
+    ) -> Union[str, bytes, Dict[str, Any]]:
         """Get data from a WFS service within a bounding box.
 
         Parameters
@@ -425,13 +378,13 @@ class WFS(WFSBase):
 
         Returns
         -------
-        Response
+        str or bytes or dict
             WFS query response within a bounding box.
         """
         utils.check_bbox(bbox)
 
         if (
-            self.version != "1.1.1"
+            self.version != "1.0.0"
             and pyproj.CRS.from_user_input(box_crs).is_geographic
             and not always_xy
         ):
@@ -447,10 +400,7 @@ class WFS(WFSBase):
             "srsName": self.crs,
         }
 
-        resp = self.session.get(self.url, payload)
-        utils.check_response(resp)
-
-        return resp
+        return ar.retrieve([self.url], self.read_method, [{"params": payload}])[0]
 
     def getfeature_bygeom(
         self,
@@ -458,7 +408,7 @@ class WFS(WFSBase):
         geo_crs: str = DEF_CRS,
         always_xy: bool = False,
         predicate: str = "INTERSECTS",
-    ) -> Response:
+    ) -> Union[str, bytes, Dict[str, Any]]:
         """Get features based on a geometry.
 
         Parameters
@@ -489,13 +439,13 @@ class WFS(WFSBase):
 
         Returns
         -------
-        Response
+        str or bytes or dict
             WFS query response based on the given geometry.
         """
-        geom = MatchCRS(geo_crs, self.crs).geometry(geometry)
+        geom = utils.match_crs(geometry, geo_crs, self.crs)
 
         if (
-            self.version != "1.1.1"
+            self.version != "1.0.0"
             and pyproj.CRS.from_user_input(geo_crs).is_geographic
             and not always_xy
         ):
@@ -524,7 +474,7 @@ class WFS(WFSBase):
         self,
         featurename: str,
         featureids: Union[List[str], str],
-    ) -> Response:
+    ) -> Union[str, bytes, Dict[str, Any]]:
         """Get features based on feature IDs.
 
         Parameters
@@ -536,7 +486,7 @@ class WFS(WFSBase):
 
         Returns
         -------
-        Response
+        str or bytes or dict
             WMS query response.
         """
         valid_features = self.get_validnames()
@@ -555,7 +505,9 @@ class WFS(WFSBase):
 
         return self.getfeature_byfilter(f"{featurename} IN ({fid_list})")
 
-    def getfeature_byfilter(self, cql_filter: str, method: str = "GET") -> Response:
+    def getfeature_byfilter(
+        self, cql_filter: str, method: str = "GET"
+    ) -> Union[str, bytes, Dict[str, Any]]:
         """Get features based on a valid CQL filter.
 
         Notes
@@ -573,7 +525,7 @@ class WFS(WFSBase):
 
         Returns
         -------
-        Response
+        str or bytes or dict
             WFS query response
         """
         if not isinstance(cql_filter, str):
@@ -594,14 +546,14 @@ class WFS(WFSBase):
         }
 
         if method == "GET":
-            resp = self.session.get(self.url, payload)
+            resp = ar.retrieve([self.url], self.read_method, [{"params": payload}])
         elif method == "POST":
             headers = {"content-type": "application/x-www-form-urlencoded"}
-            resp = self.session.post(self.url, payload, headers)
+            resp = ar.retrieve(
+                [self.url], self.read_method, [{"data": payload, "headers": headers}], "POST"
+            )
 
-        utils.check_response(resp)
-
-        return resp
+        return resp[0]
 
 
 class ServiceURL:
