@@ -1,15 +1,13 @@
 """Some utilities for PyGeoOGC."""
 import math
-import socket
-from concurrent import futures
 from dataclasses import dataclass
 from pathlib import Path
-from sqlite3 import OperationalError
-from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Tuple, Union
+from typing import Any, Dict, List, Mapping, Optional, Tuple, TypeVar, Union
 from unittest.mock import _patch, patch
 
 import defusedxml.ElementTree as etree
 import pyproj
+import shapely.geometry as sgeom
 import simplejson as json
 from requests import Response
 from requests.adapters import HTTPAdapter
@@ -17,29 +15,31 @@ from requests.exceptions import RequestException
 from requests_cache import CachedSession
 from requests_cache.backends.sqlite import DbCache
 from shapely import ops
-from shapely.geometry import (
-    LineString,
-    MultiLineString,
-    MultiPoint,
-    MultiPolygon,
-    Point,
-    Polygon,
-    box,
-)
 from urllib3 import Retry
 
-from .exceptions import InvalidInputType, ThreadingException, ZeroMatched
+from .exceptions import InvalidInputType, ServiceError
 
 DEF_CRS = "epsg:4326"
 BOX_ORD = "(west, south, east, north)"
 EXPIRE = 24 * 60 * 60
+G = TypeVar(
+    "G",
+    sgeom.Point,
+    sgeom.MultiPoint,
+    sgeom.Polygon,
+    sgeom.MultiPolygon,
+    sgeom.LineString,
+    sgeom.MultiLineString,
+    Tuple[float, float, float, float],
+    List[Tuple[float, float]],
+)
 
 
 class RetrySession:
     """Configures the passed-in session to retry on failed requests.
 
     The fails can be due to connection errors, specific HTTP response
-    codes and 30X redirections. The code is based on:
+    codes and 30X redirections. The code is was originally based on:
     https://github.com/bustawin/retry-requests
 
     Parameters
@@ -68,31 +68,21 @@ class RetrySession:
         self.cache_name = (
             Path("cache", "http_cache.sqlite") if cache_name is None else Path(cache_name)
         )
-        try:
-            backend = DbCache(self.cache_name, fast_save=True, timeout=2)
-            self.session = CachedSession(expire_after=EXPIRE, backend=backend)
-            self.session.remove_expired_responses()
-        except OperationalError:
-            self._backup_db()
+        backend = DbCache(self.cache_name, fast_save=True, timeout=1)
+        self.session = CachedSession(expire_after=EXPIRE, backend=backend)
 
-        self.retries = retries
         retry_args = {
             "total": retries,
             "read": retries,
             "connect": retries,
             "backoff_factor": backoff_factor,
             "status_forcelist": status_to_retry,
+            "allowed_methods": False,
         }
-        if hasattr(Retry.DEFAULT, "allowed_methods"):
-            retry_args.update({"allowed_methods": False})
-        else:
-            retry_args.update({"method_whitelist": False})
 
-        r = Retry(**retry_args)
-        adapter = HTTPAdapter(max_retries=r)
+        adapter = HTTPAdapter(max_retries=Retry(**retry_args))
         for prefix in prefixes:
             self.session.mount(prefix, adapter)
-        self.session.hooks = {"response": [lambda r, *args, **kwargs: r.raise_for_status()]}
 
     def get(
         self,
@@ -101,13 +91,13 @@ class RetrySession:
         headers: Optional[Mapping[str, Any]] = None,
     ) -> Response:
         """Retrieve data from a url by GET and return the Response."""
+        resp = self.session.get(url, params=payload, headers=headers)
         try:
-            return self.session.get(url, params=payload, headers=headers)
-        except OperationalError:
-            self._backup_db()
-            return self.session.get(url, params=payload, headers=headers)
-        except (ConnectionError, RequestException):
-            raise ConnectionError(f"Connection failed after {self.retries} retries.")
+            resp.raise_for_status()
+        except RequestException as ex:
+            raise ServiceError(check_response(resp.text)) from ex
+        else:
+            return resp
 
     def post(
         self,
@@ -116,93 +106,18 @@ class RetrySession:
         headers: Optional[Mapping[str, Any]] = None,
     ) -> Response:
         """Retrieve data from a url by POST and return the Response."""
+        resp = self.session.post(url, data=payload, headers=headers)
         try:
-            return self.session.post(url, data=payload, headers=headers)
-        except OperationalError:
-            self._backup_db()
-            return self.session.post(url, data=payload, headers=headers)
-        except (ConnectionError, RequestException):
-            raise ConnectionError(f"Connection failed after {self.retries} retries.")
-
-    def _backup_db(self) -> None:
-        """Use a backup database if the current database is locked."""
-        self.cache_name = Path(f"{self.cache_name.parent}/{self.cache_name.stem}_back_up.sqlite")
-        backend = DbCache(self.cache_name, fast_save=True, timeout=2)
-        self.session = CachedSession(expire_after=EXPIRE, backend=backend)
-        self.session.remove_expired_responses()
+            resp.raise_for_status()
+        except RequestException as ex:
+            raise ServiceError(check_response(resp.text)) from ex
+        else:
+            return resp
 
     @staticmethod
     def onlyipv4() -> _patch:
         """Disable IPv6 and only use IPv4."""
-        orig_getaddrinfo = socket.getaddrinfo
-
-        def getaddrinfo_ipv4(
-            host: str,
-            port: str,
-            family: socket.AddressFamily = socket.AF_INET,
-            ptype: int = 0,
-            proto: int = 0,
-            flags: int = 0,
-        ) -> List[
-            Tuple[
-                socket.AddressFamily,
-                socket.SocketKind,
-                int,
-                str,
-                Union[Tuple[str, int], Tuple[str, int, int, int]],
-            ]
-        ]:
-            return orig_getaddrinfo(
-                host=host,
-                port=port,
-                family=family,
-                type=ptype,
-                proto=proto,
-                flags=flags,
-            )
-
-        return patch("socket.getaddrinfo", side_effect=getaddrinfo_ipv4)
-
-
-def threading(
-    func: Callable,
-    iter_list: Iterable,
-    param_list: Optional[List[Any]] = None,
-    max_workers: int = 8,
-) -> List[Any]:
-    """Run a function in parallel with threading.
-
-    Notes
-    -----
-    This function is suitable for IO intensive functions.
-
-    Parameters
-    ----------
-    func : function
-        The function to be ran in threads
-    iter_list : list
-        The iterable for the function
-    param_list : list, optional
-        List of other parameters, defaults to an empty list
-    max_workers : int, optional
-        Maximum number of threads, defaults to 8
-
-    Returns
-    -------
-    list
-        A list of function returns for each iterable. The list is not ordered.
-    """
-    data = []
-    param_list = [] if param_list is None else param_list
-    with futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
-        future_to_itr = {executor.submit(func, itr, *param_list): itr for itr in iter_list}
-        for future in futures.as_completed(future_to_itr):
-            itr = future_to_itr[future]
-            try:
-                data.append(future.result())
-            except Exception as exc:  # noqa: B902
-                raise ThreadingException(itr, exc)
-    return data
+        return patch("socket.has_ipv6", False)
 
 
 def traverse_json(
@@ -223,6 +138,18 @@ def traverse_json(
     -------
     list
         The items founds in the JSON
+
+    Examples
+    --------
+    >>> from pygeoogc.utils import traverse_json
+    >>> data = [{
+    ...     "employees": [
+    ...         {"name": "Alice", "role": "dev", "nbr": 1},
+    ...         {"name": "Bob", "role": "dev", "nbr": 2}],
+    ...     "firm": {"name": "Charlie's Waffle Emporium", "location": "CA"},
+    ... },]
+    >>> traverse_json(data, ["employees", "name"])
+    [['Alice', 'Bob']]
     """
 
     def extract(
@@ -276,9 +203,9 @@ class ESRIGeomQuery:
 
     Parameters
     ----------
-    geometry : tuple or Polygon or Point or LineString
+    geometry : tuple or sgeom.Polygon or sgeom.Point or sgeom.LineString
         The input geometry which can be a point (x, y), a list of points [(x, y), ...],
-        bbox (xmin, ymin, xmax, ymax), or a Shapely's Polygon.
+        bbox (xmin, ymin, xmax, ymax), or a Shapely's sgeom.Polygon.
     wkid : int
         The Well-known ID (WKID) of the geometry's spatial reference e.g., for EPSG:4326,
         4326 should be passed. Check
@@ -290,56 +217,56 @@ class ESRIGeomQuery:
         Tuple[float, float],
         List[Tuple[float, float]],
         Tuple[float, float, float, float],
-        Polygon,
+        sgeom.Polygon,
     ]
     wkid: int
 
     def point(self) -> Dict[str, Union[str, bytes]]:
         """Query for a point."""
         if not (isinstance(self.geometry, tuple) and len(self.geometry) == 2):
-            raise InvalidInputType("geometry (point)", "tuple", "(x, y)")
+            raise InvalidInputType("geometry", "tuple", "(x, y)")
 
         geo_type = "esriGeometryPoint"
         geo_json = dict(zip(("x", "y"), self.geometry))
-        return self.get_payload(geo_type, geo_json)
+        return self._get_payload(geo_type, geo_json)
 
     def multipoint(self) -> Dict[str, Union[str, bytes]]:
         """Query for a multi-point."""
         if not (isinstance(self.geometry, list) and all(len(g) == 2 for g in self.geometry)):
-            raise InvalidInputType("geometry (multi-point)", "list of tuples", "[(x, y), ...]")
+            raise InvalidInputType("geometry", "list of tuples", "[(x, y), ...]")
 
         geo_type = "esriGeometryMultipoint"
         geo_json = {"points": [[x, y] for x, y in self.geometry]}
-        return self.get_payload(geo_type, geo_json)
+        return self._get_payload(geo_type, geo_json)
 
     def bbox(self) -> Dict[str, Union[str, bytes]]:
         """Query for a bbox."""
         if not (isinstance(self.geometry, (tuple, list)) and len(self.geometry) == 4):
-            raise InvalidInputType("geometry (bbox)", "tuple or list", BOX_ORD)
+            raise InvalidInputType("geometry", "tuple", BOX_ORD)
 
         geo_type = "esriGeometryEnvelope"
         geo_json = dict(zip(("xmin", "ymin", "xmax", "ymax"), self.geometry))
-        return self.get_payload(geo_type, geo_json)
+        return self._get_payload(geo_type, geo_json)
 
     def polygon(self) -> Dict[str, Union[str, bytes]]:
         """Query for a polygon."""
-        if not isinstance(self.geometry, Polygon):
-            raise InvalidInputType("geometry", "Shapely's Polygon")
+        if not isinstance(self.geometry, sgeom.Polygon):
+            raise InvalidInputType("geometry", "Polygon")
 
         geo_type = "esriGeometryPolygon"
         geo_json = {"rings": [[[x, y] for x, y in zip(*self.geometry.exterior.coords.xy)]]}
-        return self.get_payload(geo_type, geo_json)
+        return self._get_payload(geo_type, geo_json)
 
     def polyline(self) -> Dict[str, Union[str, bytes]]:
         """Query for a polyline."""
-        if not isinstance(self.geometry, LineString):
-            raise InvalidInputType("geometry", "Shapely's LineString")
+        if not isinstance(self.geometry, sgeom.LineString):
+            raise InvalidInputType("geometry", "LineString")
 
         geo_type = "esriGeometryPolyline"
         geo_json = {"paths": [[[x, y] for x, y in zip(*self.geometry.coords.xy)]]}
-        return self.get_payload(geo_type, geo_json)
+        return self._get_payload(geo_type, geo_json)
 
-    def get_payload(self, geo_type: str, geo_json: Dict[str, Any]) -> Dict[str, Union[str, bytes]]:
+    def _get_payload(self, geo_type: str, geo_json: Dict[str, Any]) -> Dict[str, Union[str, bytes]]:
         """Generate a request payload based on ESRI template.
 
         Parameters
@@ -362,100 +289,65 @@ class ESRIGeomQuery:
         }
 
 
-class MatchCRS:
+def match_crs(geom: G, in_crs: str, out_crs: str) -> G:
     """Reproject a geometry to another CRS.
 
     Parameters
     ----------
+    geom : list or tuple or geometry
+        Input geometry which could be a list of coordinates such as ``[(x1, y1), ...]``,
+        a bounding box like so ``(xmin, ymin, xmax, ymax)``, or any valid ``shapely``'s
+        geometry such as sgeom.Polygon, sgeom.MultiPolygon, etc..
     in_crs : str
         Spatial reference of the input geometry
     out_crs : str
         Target spatial reference
+
+    Returns
+    -------
+    sgeom.LineString, sgeom.MultiLineString, sgeom.Polygon, sgeom.MultiPolygon, sgeom.Point, or sgeom.MultiPoint
+        Input geometry in the specified CRS.
+
+    Examples
+    --------
+    >>> from pygeoogc.utils import match_crs
+    >>> from shapely.geometry import Point
+    >>> point = Point(-7766049.665, 5691929.739)
+    >>> match_crs(point, "epsg:3857", "epsg:4326").xy
+    (array('d', [-69.7636111130079]), array('d', [45.44549114818127]))
+    >>> bbox = (-7766049.665, 5691929.739, -7763049.665, 5696929.739)
+    >>> match_crs(bbox, "epsg:3857", "epsg:4326")
+    (-69.7636111130079, 45.44549114818127, -69.73666165448431, 45.47699468552394)
+    >>> coords = [(-7766049.665, 5691929.739)]
+    >>> match_crs(coords, "epsg:3857", "epsg:4326")
+    [(-69.7636111130079, 45.44549114818127)]
     """
+    project = pyproj.Transformer.from_crs(in_crs, out_crs, always_xy=True).transform
 
-    def __init__(self, in_crs: str, out_crs: str):
-        self.project = pyproj.Transformer.from_crs(in_crs, out_crs, always_xy=True).transform
+    if isinstance(
+        geom,
+        (
+            sgeom.Polygon,
+            sgeom.LineString,
+            sgeom.MultiLineString,
+            sgeom.MultiPolygon,
+            sgeom.Point,
+            sgeom.MultiPoint,
+        ),
+    ):
+        return ops.transform(project, geom)
 
-    def geometry(
-        self, geom: Union[Polygon, LineString, MultiLineString, MultiPolygon, Point, MultiPoint]
-    ) -> Union[Polygon, LineString, MultiLineString, MultiPolygon, Point, MultiPoint]:
-        """Reproject a geometry to the specified output CRS.
+    if isinstance(geom, tuple) and len(geom) == 4:
+        return ops.transform(project, sgeom.box(*geom)).bounds
 
-        Parameters
-        ----------
-        geom : LineString, MultiLineString, Polygon, MultiPolygon, Point, or MultiPoint
-            Input geometry.
+    if isinstance(geom, list) and all(len(c) == 2 for c in geom):
+        return list(zip(*project(*zip(*geom))))
 
-        Returns
-        -------
-        LineString, MultiLineString, Polygon, MultiPolygon, Point, or MultiPoint
-            Input geometry in the specified CRS.
-
-        Examples
-        --------
-        >>> from pygeoogc import MatchCRS
-        >>> from shapely.geometry import Point
-        >>> point = Point(-7766049.665, 5691929.739)
-        >>> MatchCRS("epsg:3857", "epsg:4326").geometry(point).xy
-        (array('d', [-69.7636111130079]), array('d', [45.44549114818127]))
-        """
-        if not isinstance(
-            geom, (Polygon, LineString, MultiLineString, MultiPolygon, Point, MultiPoint)
-        ):
-            types = "LineString, MultiLineString, Polygon, MultiPolygon, Point, or MultiPoint"
-            raise InvalidInputType("geom", types)
-
-        return ops.transform(self.project, geom)
-
-    def bounds(self, geom: Tuple[float, float, float, float]) -> Tuple[float, float, float, float]:
-        """Reproject a bounding box to the specified output CRS.
-
-        Parameters
-        ----------
-        geom : tuple
-            Input bounding box (xmin, ymin, xmax, ymax).
-
-        Returns
-        -------
-        tuple
-            Input bounding box in the specified CRS.
-
-        Examples
-        --------
-        >>> from pygeoogc import MatchCRS
-        >>> bbox = (-7766049.665, 5691929.739, -7763049.665, 5696929.739)
-        >>> MatchCRS("epsg:3857", "epsg:4326").bounds(bbox)
-        (-69.7636111130079, 45.44549114818127, -69.73666165448431, 45.47699468552394)
-        """
-        if not (isinstance(geom, tuple) and len(geom) == 4):
-            raise InvalidInputType("geom", "tuple", BOX_ORD)
-
-        return ops.transform(self.project, box(*geom)).bounds
-
-    def coords(self, geom: List[Tuple[float, float]]) -> List[Tuple[Any, ...]]:
-        """Reproject a list of coordinates to the specified output CRS.
-
-        Parameters
-        ----------
-        geom : list of tuple
-            Input coords [(x1, y1), ...].
-
-        Returns
-        -------
-        tuple
-            Input list of coords in the specified CRS.
-
-        Examples
-        --------
-        >>> from pygeoogc import MatchCRS
-        >>> coords = [(-7766049.665, 5691929.739)]
-        >>> MatchCRS("epsg:3857", "epsg:4326").coords(coords)
-        [(-69.7636111130079, 45.44549114818127)]
-        """
-        if not (isinstance(geom, list) and all(len(c) == 2 for c in geom)):
-            raise InvalidInputType("geom", "list of tuples", "[(x1, y1), ...]")
-
-        return list(zip(*self.project(*zip(*geom))))
+    gtypes = (
+        "a list of coordinates such as [(x1, y1), ...],"
+        + "a bounding box like so (xmin, ymin, xmax, ymax), or any valid shapely's geometry."
+    )
+    raise InvalidInputType("geom", gtypes)
 
 
 def check_bbox(bbox: Tuple[float, float, float, float]) -> None:
@@ -485,14 +377,14 @@ def bbox_resolution(
     """
     check_bbox(bbox)
 
-    bbox = MatchCRS(bbox_crs, DEF_CRS).bounds(bbox)
+    bbox = match_crs(bbox, bbox_crs, DEF_CRS)
     west, south, east, north = bbox
     geod = pyproj.Geod(ellps="WGS84")
 
-    linex = LineString([Point(west, south), Point(east, south)])
+    linex = sgeom.LineString([sgeom.Point(west, south), sgeom.Point(east, south)])
     delx = geod.geometry_length(linex)
 
-    liney = LineString([Point(west, south), Point(west, north)])
+    liney = sgeom.LineString([sgeom.Point(west, south), sgeom.Point(west, north)])
     dely = geod.geometry_length(liney)
 
     return int(delx / resolution), int(dely / resolution)
@@ -504,7 +396,7 @@ def bbox_decompose(
     box_crs: str = DEF_CRS,
     max_px: int = 8000000,
 ) -> List[Tuple[Tuple[float, float, float, float], str, int, int]]:
-    """Split the bounding box vertically for WMS requests.
+    r"""Split the bounding box vertically for WMS requests.
 
     Parameters
     ----------
@@ -520,11 +412,29 @@ def bbox_decompose(
 
     Returns
     -------
-    tuple
-        The first element is a list of bboxes and the second one is width of the last bbox
+    list of tuples
+        Each tuple includes the following elements:
+
+        * Tuple of length 4 that represents a bounding box (west, south, east, north) of a cell,
+        * A label that represents cell ID starting from bottom-left to top-right, for example a
+          2x2 decomposition has the following labels::
+
+          |---------|---------|
+          |         |         |
+          |   0_1   |   1_1   |
+          |         |         |
+          |---------|---------|
+          |         |         |
+          |   0_0   |   1_0   |
+          |         |         |
+          |---------|---------|
+
+        * Raster width of a cell,
+        * Raster height of a cell.
+
     """
     check_bbox(bbox)
-    _bbox = MatchCRS(box_crs, DEF_CRS).bounds(bbox)
+    _bbox = match_crs(bbox, box_crs, DEF_CRS)
     width, height = bbox_resolution(_bbox, resolution, DEF_CRS)
 
     n_px = width * height
@@ -573,13 +483,19 @@ def bbox_decompose(
     bboxs = []
     for i, ((bottom, top), h) in enumerate(zip(lats, heights)):
         for j, ((left, right), w) in enumerate(zip(lons, widths)):
-            bx_crs = MatchCRS(DEF_CRS, box_crs).bounds((left, bottom, right, top))
+            bx_crs = match_crs((left, bottom, right, top), DEF_CRS, box_crs)
             bboxs.append((bx_crs, f"{i}_{j}", w, h))
     return bboxs
 
 
-def check_response(resp: Response) -> None:
-    """Check if a ``requests.Resonse`` returned an error message."""
-    if resp.headers["Content-Type"] == "application/xml":
-        root = etree.fromstring(resp.text)
-        raise ZeroMatched(root[0][0].text.strip())
+def check_response(resp: str) -> str:
+    """Extract error message from a response, if any."""
+    try:
+        root = etree.fromstring(resp)
+    except etree.ParseError:
+        return resp
+    else:
+        try:
+            return root[-1][0].text.strip()
+        except IndexError:
+            return root[-1].text.strip()
