@@ -2,15 +2,16 @@
 import io
 import sys
 import zipfile
+from unittest.mock import patch
 
 import pytest
 from shapely.geometry import LineString, Polygon
 
 import pygeoogc as ogc
-from pygeoogc import WFS, WMS, ArcGISRESTful, MatchCRS, RetrySession, ServiceURL, utils
+from pygeoogc import WFS, WMS, ArcGISRESTful, RetrySession, ServiceURL, utils
 
 DEF_CRS = "epsg:4326"
-ALT_CRS = "epsg:2149"
+ALT_CRS = "epsg:4269"
 GEO_NAT = Polygon(
     [[-69.77, 45.07], [-69.31, 45.07], [-69.31, 45.45], [-69.77, 45.45], [-69.77, 45.07]]
 )
@@ -29,8 +30,8 @@ COORDS = [(-118.72, 34.118), (-118.31, 34.518)]
 class TestREST:
     wbd_url: str = ServiceURL().restful.wbd
     fab_url: str = f"{ServiceURL().restful.nhd_fabric}/1"
-    epa_url: str = "https://watersgeo.epa.gov/arcgis/rest/services/NHDPlus/NHDPlus/MapServer"
-    nhd_url: str = ServiceURL().restful.nhdplushr_edits
+    epa_url: str = ServiceURL().restful.nhdplus_epa
+    nhd_url: str = ServiceURL().restful.nhdplushr
 
     def test_byid(self):
         """RESTFul by ID"""
@@ -76,16 +77,41 @@ class TestREST:
 
         service = ArcGISRESTful(self.epa_url, 2, outformat="json")
         service.oids_bygeom(
-            geom, geo_crs="epsg:4269", sql_clause="FTYPE NOT IN (420,428,566)", distance=1500
+            geom, geo_crs=ALT_CRS, sql_clause="FTYPE NOT IN (420,428,566)", distance=1500
         )
         resp = service.get_features(return_m=True)
         assert len(resp[0]["features"]) == 3
 
+    def test_retry(self):
+        rest_url = "/".join(
+            [
+                "http://cohgims.houstontx.gov/arcgis/rest/services",
+                "PUD_Utility/StormwaterUtilities/MapServer/3",
+            ]
+        )
+        rest = ArcGISRESTful(rest_url)
+        print(rest)
+        oids = [
+            1006322,
+            1006323,
+            1006324,
+            1245206,
+            323351,
+            324961,
+            474077,
+            503848,
+            584849,
+            585138,
+        ]
+        rest.featureids = rest.partition_oids(oids)
+        resp = rest.get_features()
+        assert len(resp) == 3
+
     @pytest.mark.slow
     def test_bysql(self):
         """RESTFul by SQL filter"""
         hr = ArcGISRESTful(self.nhd_url, 2, outformat="json")
-        hr.oids_bysql("NHDFlowline.PERMANENT_IDENTIFIER IN ('103455178', '103454362', '103453218')")
+        hr.oids_bysql("PERMANENT_IDENTIFIER IN ('103455178', '103454362', '103453218')")
         resp = hr.get_features(return_m=True)
 
         assert len(resp[0]["features"]) == 3
@@ -94,20 +120,21 @@ class TestREST:
     def test_byfield(self):
         """RESTFul by SQL filter"""
         hr = ArcGISRESTful(self.nhd_url, 2, outformat="json")
-        hr.oids_byfield("NHDFlowline.PERMANENT_IDENTIFIER", ["103455178", "103454362", "103453218"])
+        hr.oids_byfield("PERMANENT_IDENTIFIER", ["103455178", "103454362", "103453218"])
         resp = hr.get_features()
 
         assert len(resp[0]["features"]) == 3
 
 
 class TestWMS:
-    wms_url: str = ServiceURL().wms.fws
+    wms_url: str = ServiceURL().wms.gebco
+    layer: str = "GEBCO_LATEST"
 
     def test_v111(self):
         """WMS version 1.1.1"""
         wms = WMS(
             self.wms_url,
-            layers="0",
+            layers=self.layer,
             outformat="image/tiff",
             crs=DEF_CRS,
             version="1.1.1",
@@ -115,16 +142,16 @@ class TestWMS:
         )
         r_dict = wms.getmap_bybox(GEO_NAT.bounds, 20, DEF_CRS)
         assert (
-            wms.get_validlayers()["0"] == "Wetlands_Raster"
-            and sys.getsizeof(r_dict["0_dd_0_0"]) == 12536763
+            wms.get_validlayers()[self.layer] == self.layer
+            and sys.getsizeof(r_dict[f"{self.layer}_dd_0_0"]) == 11485083
         )
 
     def test_bybox(self):
         """WMS by bounding box"""
-        wms = WMS(self.wms_url, layers="0", outformat="image/tiff", crs=DEF_CRS)
+        wms = WMS(self.wms_url, layers=self.layer, outformat="image/tiff", crs=DEF_CRS)
         print(wms)
-        r_dict = wms.getmap_bybox(GEO_NAT.bounds, 20, DEF_CRS)
-        assert sys.getsizeof(r_dict["0_dd_0_0"]) == 12536763
+        r_dict = wms.getmap_bybox(GEO_NAT.bounds, 20, DEF_CRS, max_px=3e6)
+        assert sum(sys.getsizeof(r) for r in r_dict.values()) == 11495526
 
 
 class TestWFS:
@@ -133,30 +160,42 @@ class TestWFS:
         layer="wmadata:gagesii",
         outformat="application/json",
         version="2.0.0",
-        crs="epsg:4269",
+        crs=ALT_CRS,
     )
 
     def test_byid(self):
         """WFS by ID"""
         print(self.wfs)
         st = self.wfs.getfeature_byid("staid", "01031500")
-        assert st.json()["numberMatched"] == 1
+        assert st["numberMatched"] == 1
 
     def test_bygeom(self):
         """WFS by geometry"""
         r = self.wfs.getfeature_bygeom(GEO_URB, geo_crs=DEF_CRS, always_xy=False)
-        assert len(r.json()["features"]) == 7
+        assert len(r["features"]) == 7
 
     def test_bybox(self):
         """WFS by bounding box"""
         bbox = GEO_URB.bounds
         r = self.wfs.getfeature_bybox(bbox, box_crs=DEF_CRS, always_xy=True)
-        assert len(r.json()["features"]) == 7
+        assert len(r["features"]) == 7
 
     def test_byfilter(self):
         """WFS by CQL filter"""
         wb = self.wfs.getfeature_byfilter("staid LIKE '010315%'")
-        assert len(utils.traverse_json(wb.json(), ["features", "geometry", "coordinates"])) == 2
+        assert len(utils.traverse_json(wb, ["features", "geometry", "coordinates"])) == 2
+
+    def test_wfs111(self):
+        """WFS 1.0.0 by geom"""
+        wfs = WFS(
+            ServiceURL().wfs.waterdata,
+            layer="wmadata:gagesii",
+            outformat="json",
+            version="1.0.0",
+            crs=ALT_CRS,
+        )
+        r = wfs.getfeature_bygeom(GEO_URB, geo_crs=DEF_CRS, always_xy=False)
+        assert len(r["features"]) == 7
 
 
 def test_decompose():
@@ -175,8 +214,7 @@ def test_decompose():
 )
 def test_matchcrs(geo, gtype, expected):
     """Match CRS"""
-    match_crs = MatchCRS(DEF_CRS, ALT_CRS)
-    matched = getattr(match_crs, gtype)(geo)
+    matched = utils.match_crs(geo, DEF_CRS, "epsg:2149")
     if gtype == "coords":
         val = matched[-1][0]
     elif gtype == "geometry":
@@ -198,17 +236,14 @@ def test_esriquery():
 
 def test_ipv4():
     """Only IPv4"""
-    url = (
-        "https://edcintl.cr.usgs.gov/downloads/sciweb1/shared/uswem/web/conus"
-        + "/eta/modis_eta/daily/downloads/det2004003.modisSSEBopETactual.zip"
-    )
+    url = f"{ServiceURL().http.ssebopeta}/det2005150.modisSSEBopETactual.zip"
     session = RetrySession()
-    with session.onlyipv4():
+    with patch("socket.has_ipv6", False):
         r = session.get(url)
         z = zipfile.ZipFile(io.BytesIO(r.content))
         fname = z.read(z.filelist[0].filename)
 
-    assert sys.getsizeof(fname) == 4361682
+    assert sys.getsizeof(fname) == 20283088
 
 
 def test_urls():
